"""Created on Aug 10 23:35:38 2024"""

from .generators import (
    multiple_models,
    multi_exponential,
    multi_base,
    multi_gaussian,
    multi_laplace,
    multi_gamma,
    multi_log_normal,
    multi_skew_normal,
    multi_half_normal,
    multi_folded_normal,
    multi_chi_squared,
)
